/**
 * @file okc_comm.c
 *
 * @brief Main file of libOpenKC. 
 *
 * All basic communication from/to robot is handled here.
 *
 * @author Matthias Schöpfer
 * @version 0.6
 * @date 1.7.2009
 */

#include "okc_comm.h"
/* #define BENCH 1 */
#define TICKS_DEBUG 1
#ifdef BENCH
#include <sys/time.h>
#endif

/* #define DEBUG_CONNECTION 1 */

/**
 * Helper function intern to OpenKC. Parses XML-Section related to coords-set.
 *
 * @param node #xmlNode that holds coords-data (X,Y,Z,A,B,C).
 *
 * @param gen_coords #coords to be filled by function.
 *
 * @return EXIT_SUCCESS on success, EXIT_FAILURE on failure.
 */
static int 
_parse_coords_node (xmlNode* node, coords* gen_coords)
{
  char* buffer;
  char* detect_error;

  detect_error = buffer = (char*) xmlGetProp (node,BAD_CAST "X");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property X where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }

  gen_coords->x = strtod (buffer, &detect_error);
  
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_coords, Property X\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "Y");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property Y where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }

  gen_coords->y = strtod (buffer,&detect_error);
  
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_coords, Property Y\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "Z");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property Z where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }

  gen_coords->z = strtod (buffer,&detect_error);
  
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_coords, Property Z\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property A where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }

  gen_coords->a = strtod (buffer,&detect_error);

  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_coords, Property A\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "B");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property B where it "
	       "should be!\n");
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }

  gen_coords->b = strtod (buffer,&detect_error);

  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_coords, Property B\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "C");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property C where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }

  gen_coords->c = strtod (buffer,&detect_error);

  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_coords, Property C\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  return (EXIT_SUCCESS);
}





/**
 * Helper function intern to OpenKC. Parses XML-Section related to long-int-set.
 *
 * @param node #xmlNode that holds long int data.
 *
 * @param llip pointer to long long int to be filled.
 *
 * @return EXIT_SUCCESS on success, EXIT_FAILURE on failure.
 */
static int
_parse_text_to_long_long_int_node (xmlNode* node, long long int* llip)
{
  unsigned char* buffer = NULL;
  unsigned char* detect_error = NULL;

  int outlen, inlen;
  int ret;

  outlen = inlen = xmlUTF8Strlen (node->children->content);

  buffer = malloc ((inlen+1)*sizeof (char));

  buffer[inlen] = '\000';

  ret = UTF8Toisolat1 (buffer,&outlen,node->children->content, 
		 &inlen);
  
  (*llip) = strtoll ((char*) buffer, (char**) &detect_error, 10);

  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in IPOC\n",buffer);
      return (EXIT_FAILURE);
    }
  free (buffer);
  return (EXIT_SUCCESS);
}

/**
 * Helper function intern to OpenKC. Parses XML-Section related to int-set.
 *
 * @param node #xmlNode that holds int data.
 *
 * @param ip pointer to int to be filled.
 *
 * @return EXIT_SUCCESS on success, EXIT_FAILURE on failure.
 */
static int
_parse_text_to_int_node (xmlNode* node, int* ip)
{
  unsigned char* buffer = NULL;
  unsigned char* detect_error = NULL;

  int outlen, inlen;
  int ret;

  outlen = inlen = xmlUTF8Strlen (node->children->content);

  buffer = malloc ((inlen+1)*sizeof (char));

  buffer[inlen] = '\000';

  ret = UTF8Toisolat1 (buffer,&outlen,node->children->content, 
		 &inlen);
  
  (*ip) = (int) strtol ((char*) buffer, (char**) &detect_error, 10);

  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in status\n",buffer);
      return (EXIT_FAILURE);
    }
  free (buffer);
  return (EXIT_SUCCESS);
}

/**
 * Helper function intern to OpenKC. Parses XML-Section related to lbr_axis-set.
 *
 * @param node #xmlNode that holds lbr_axis (A1,..,A6)/(E1,..,E6) data.
 *
 * @param gen_axis pointer to #lbr_axis to be filled.
 *
 * @return EXIT_SUCCESS on success, EXIT_FAILURE on failure.
 */

  
static int
_parse_axis_node (xmlNode* node, lbr_axis* gen_axis)
{
  char* buffer;
  char* detect_error;
  
  if ((NULL == (buffer = (char*) xmlGetProp (node, BAD_CAST "A1"))) && (NULL == (detect_error = (char*) xmlGetProp (node, BAD_CAST "E1"))))
    {
      fprintf (stderr,"parse_tree: wrong node?!\n");
      return (EXIT_FAILURE); 
    }

  if (NULL != buffer)
    {
      xmlFree (buffer);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A1");

      if (NULL == buffer)
	{
	  fprintf (stderr,"parse_tree: Did not find Property A1 where it "
		   "should be!\n");
	  return (EXIT_FAILURE);
	}
      
      gen_axis->a1 = strtod (buffer,&detect_error);
      
      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property A1\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A2");

      if (NULL == buffer)
	{
	  fprintf (stderr,"parse_tree: Did not find Property A2 where it "
		   "should be!\n");
	  return (EXIT_FAILURE);
	}

      gen_axis->a2 = strtod (buffer,&detect_error);
  
      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property A2\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A3");

      if (NULL == buffer)
	{
	  fprintf (stderr,"parse_tree: Did not find Property A3 where it "
		   "should be!\n");

	  return (EXIT_FAILURE);
	}

      gen_axis->a3 = strtod (buffer,&detect_error);
  
      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property A3\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A4");

      if (NULL == buffer)
	{
	  fprintf (stderr,"parse_tree: Did not find Property A4 where it "
		   "should be!\n");
	  return (EXIT_FAILURE);
	}
      
      gen_axis->a4 = strtod (buffer,&detect_error);

      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property A4\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A5");

      if (NULL == buffer)
	{
	  fprintf (stderr,"parse_tree: Did not find Property A5 where it "
		   "should be!\n");
	  return (EXIT_FAILURE);
	}
      
      gen_axis->a5 = strtod (buffer,&detect_error);

      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property A5\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A6");

      if (NULL == buffer)
	{
	  fprintf (stderr,"parse_tree: Did not find Property A6 where it "
		   "should be!\n");
	  return (EXIT_FAILURE);
	}

      gen_axis->a6 = strtod (buffer,&detect_error);

      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property A6\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
    }
  else
    {
      xmlFree (detect_error);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "E1");

      if (NULL == buffer)
	{
	  fprintf (stderr,"Did not find Property E1 where it should be!\n");
	  return (EXIT_FAILURE);
	}
		 
      gen_axis->e1 = strtod (buffer,&detect_error);

      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property E1\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
    }
  return (EXIT_SUCCESS);
}
/**
 * Helper function intern to OpenKC. Parses XML-Section related to lbr_axis-set.
 * Parses Axis-Torques.
 *
 * @param node #xmlNode that holds Torque-Data (A1,...A6,E1) data.
 *
 * @param gen_axis pointer to #lbr_axis to be filled.
 *
 * @return EXIT_SUCCESS on success, EXIT_FAILURE on failure.
 */
static int
_parse_torque_node (xmlNode* node, lbr_axis* gen_axis)
{
  char* buffer=NULL;
  char* detect_error=NULL;
  
  if ((NULL == (buffer = (char*) xmlGetProp (node, BAD_CAST "A1"))) && (NULL == (detect_error = (char*) xmlGetProp (node, BAD_CAST "E1"))))
    {
      fprintf (stderr,"parse_tree: wrong node?!\n");
      return (EXIT_FAILURE); 
    }
  if (NULL != buffer)
    xmlFree (buffer);
  if (NULL != detect_error)
    xmlFree (detect_error);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A1");
  
  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property A1 where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }
      
  gen_axis->a1 = strtod (buffer,&detect_error);
  
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property A1\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);  
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A2");
  
  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property A2 where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }
  
  gen_axis->a2 = strtod (buffer,&detect_error);
  
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property A2\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A3");
  
  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property A3 where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }
  
  gen_axis->a3 = strtod (buffer,&detect_error);
  
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property A3\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A4");
  
  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property A4 where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }
  
  gen_axis->a4 = strtod (buffer,&detect_error);
  
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property A4\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A5");
  
  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property A5 where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }
  
  gen_axis->a5 = strtod (buffer,&detect_error);
  
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property A5\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A6");
  
  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property A6 where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }
  
  gen_axis->a6 = strtod (buffer,&detect_error);
  
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property A6\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "E1");
  
  if (NULL == buffer)
    {
      fprintf (stderr,"Did not find Property E1 where it should be!\n");
      return (EXIT_FAILURE);
    }
  
  gen_axis->e1 = strtod (buffer,&detect_error);
  
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property E1\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  return (EXIT_SUCCESS);
}

/**
 * Helper function intern to OpenKC. Parses XML-Input-Tree. Uses intern 
 * helper functions for that.
 *
 * @param doc Pointer to xmlDoc holding the XML-Input from robot.
 * @param RIst Pointer to #coords, will be filled with actual position 
 *        ($pos_act) from robot.
 * @param IPos Pointer to #lbr_axis, will be filles with actual axis
 *        joint values from robot.
 * @param ITorque Pointer to #lbr_axis, will be filled with actual torques
 *        from the robot joints.
 * @param IFT Pointer to #coords, will be filled with cartesian forces and
 *        torques. Values are ESTIMATED using $TORQUE_TCP_EST
 * @param status Pointer to int, will hold robots "status" flag.
 * @param ipoc Pointer to long long int, will be filled with the robots IPOC
 *        value, which has to be set to answer robot correctly.
 *
 * @return EXIT_SUCCESS on success, EXIT_FAILURE on failure.
 */

static int
_parse_tree (xmlDoc* doc, coords* RIst, coords* RSol, lbr_axis* IPos, 
	     lbr_axis* SPos, lbr_axis* ITorque, coords* IFT, int* status, 
	     long long int* ipoc)
{
  xmlNode* root_node = NULL;
  xmlNode* cur_node = NULL;

  
  root_node = xmlDocGetRootElement(doc);
  
  if (NULL == root_node)
    {
      fprintf (stderr,"_parse_tree: Error parsing doctree!\n");
      return (EXIT_FAILURE);
    }

  if (0 != xmlStrcmp (root_node->name,BAD_CAST "Rob"))
    return (EXIT_FAILURE);

  cur_node = root_node->xmlChildrenNode;
  
  while (NULL != cur_node)
    {
      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name, BAD_CAST "RIst"))) 
	{
	  if (EXIT_SUCCESS !=  _parse_coords_node (cur_node, RIst))
	    return (EXIT_FAILURE);
	}

      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name, BAD_CAST "RSol"))) 
	{
	  if (EXIT_SUCCESS !=  _parse_coords_node (cur_node, RSol))
	    return (EXIT_FAILURE);
	}
	  
      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name,BAD_CAST "AIPos")))
	{
	  if (EXIT_SUCCESS != _parse_axis_node (cur_node, IPos))
	    return (EXIT_FAILURE);
	}
      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name,BAD_CAST "EIPos")))
	{
	  if (EXIT_SUCCESS != _parse_axis_node (cur_node, IPos))
	    return (EXIT_FAILURE);
	}
      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name,BAD_CAST "ASPos")))
	{
	  if (EXIT_SUCCESS != _parse_axis_node (cur_node, SPos))
	    return (EXIT_FAILURE);
	}
      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name,BAD_CAST "ESPos")))
	{
	  if (EXIT_SUCCESS != _parse_axis_node (cur_node, SPos))
	    return (EXIT_FAILURE);
	}
      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name,BAD_CAST "Torques")))
	{
	  if (EXIT_SUCCESS != _parse_torque_node (cur_node, ITorque))
	    return (EXIT_FAILURE);
	  if (EXIT_SUCCESS != _parse_coords_node (cur_node, IFT))
	    return (EXIT_FAILURE);
/* 	  if (EXIT_SUCCESS != _parse_variance_node (cur_node, IVar)) */
/* 	    return (EXIT_FAILURE); */
	}
      if ((cur_node->type == XML_ELEMENT_NODE) &&
	  (0 == xmlStrcmp (cur_node->name, BAD_CAST "Status")))
	{
	  if (EXIT_SUCCESS != _parse_text_to_int_node (cur_node, status))
	    return (EXIT_FAILURE);
	}
      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name,BAD_CAST "IPOC")))
	{
	  if (EXIT_SUCCESS != _parse_text_to_long_long_int_node (cur_node, ipoc))
	    return (EXIT_FAILURE);
	}
      cur_node = cur_node->next;
    }
  return (EXIT_SUCCESS);
}
	      
	  
	  
/**
 * Helper function intern to OpenKC. Creates a XML-Node holding the new 
 * cartesian corrections that will be send to the robot.
 *
 * @param node #xmlNode that will be filled with rkorr.
 *
 * @param rkorr pointer to #coords to be copied to node.
 *
 * @return Pointer to new #xmlNode
 */  
  
  

static xmlNode*
_write_rkorr_to_node (xmlNode* node, coords rkorr)
{

  char attr[OKC_ARBITRARY_STRING_LENGTH];

//  printf("!!! rkorr.x: |%f| \n",rkorr.x);
//  printf("!!! rkorr.y: |%f| \n",rkorr.y);
//  printf("!!! rkorr.z: |%f| \n",rkorr.z);
//
//  printf("!!! rkorr.x: |%f| \n",rkorr.x);

  int ret = snprintf (attr,OKC_ARBITRARY_STRING_LENGTH, "%1.3f",rkorr.x);

//  printf("!!! ret rkorr.x conv: %d \n",ret);
//  printf("!!! char array: |%s| \n", attr);

  xmlNewProp (node,BAD_CAST "X",BAD_CAST attr);
  
//  printf("!!! rkorr.y: |%f| \n",rkorr.y);

  ret = snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",rkorr.y);

//  printf("!!! ret rkorr.y conv: %d \n",ret);
//  printf("!!! char array: |%s| \n", attr);

  
  xmlNewProp (node,BAD_CAST "Y",BAD_CAST attr);
  
//  printf("!!! rkorr.z: |%f| \n",rkorr.z);
  
  ret = snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",rkorr.z);

//  printf("!!! ret rkorr.z conv: %d \n",ret);
//  printf("!!! char array: |%s| \n", attr);

  //exit(-1);

  xmlNewProp (node,BAD_CAST "Z",BAD_CAST attr);

  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",rkorr.a);
  
  xmlNewProp (node,BAD_CAST "A",BAD_CAST attr);

  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",rkorr.b);
  
  xmlNewProp (node,BAD_CAST "B",BAD_CAST attr);

  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",rkorr.c);
  
  xmlNewProp (node,BAD_CAST "C",BAD_CAST attr);

  return node;
}

/**
 * Helper function intern to OpenKC. Creates a XML-Node holding the new 
 * axis corrections that will be send to the robot.
 *
 * @param node #xmlNode that will be filled with akorr.
 *
 * @param akorr pointer to #lbr_axis to be copied to node.
 *
 * @return Pointer to new #xmlNode
 */  

static xmlNode*
_write_akorr_to_node (xmlNode* node, lbr_axis akorr)
{

  char attr[OKC_ARBITRARY_STRING_LENGTH];

  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH, "%1.3f",akorr.a1);

  xmlNewProp (node,BAD_CAST "A1",BAD_CAST attr);
  
  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",akorr.a2);
  
  xmlNewProp (node,BAD_CAST "A2",BAD_CAST attr);
  
  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",akorr.a3);
  
  xmlNewProp (node,BAD_CAST "A3",BAD_CAST attr);

  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",akorr.a4);
  
  xmlNewProp (node,BAD_CAST "A4",BAD_CAST attr);

  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",akorr.a5);
  
  xmlNewProp (node,BAD_CAST "A5",BAD_CAST attr);

  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",akorr.a6);
  
  xmlNewProp (node,BAD_CAST "A6",BAD_CAST attr);

  return node;
}

/**
 * Helper function intern to OpenKC. Creates a XML-Node holding the e1 
 * axis correction that will be send to the robot.
 *
 * @param node #xmlNode that will be filled with akorr.
 *
 * @param akorr pointer to #lbr_axis to be copied to node.
 *
 * @return Pointer to new #xmlNode
 */  

static xmlNode*
_write_ekorr_to_node (xmlNode* node, lbr_axis akorr)
{
  char attr[OKC_ARBITRARY_STRING_LENGTH];
  
  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",akorr.e1);
  
  xmlNewProp (node, BAD_CAST "E1", BAD_CAST attr);

  return node;
}

/**
 * Helper function intern to OpenKC. Creates a XML-Node holding the new 
 * params (for impedance control) that will be send to the robot.
 *
 * @param node #xmlNode that will be filled with akorr.
 *
 * @param params pointer to double array of length 7 (even, if only 6 
 *        parameters are used!).
 *
 * @return Pointer to new #xmlNode
 */  

static xmlNode*
_write_params_to_node (xmlNode* node, double* params)
{

  char attr[OKC_ARBITRARY_STRING_LENGTH];

  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH, "%1.3f",params[0]);

  xmlNewProp (node,BAD_CAST "P1",BAD_CAST attr);
  
  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",params[1]);
  
  xmlNewProp (node,BAD_CAST "P2",BAD_CAST attr);
  
  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",params[2]);
  
  xmlNewProp (node,BAD_CAST "P3",BAD_CAST attr);

  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",params[3]);
  
  xmlNewProp (node,BAD_CAST "P4",BAD_CAST attr);

  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",params[4]);
  
  xmlNewProp (node,BAD_CAST "P5",BAD_CAST attr);

  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",params[5]);
  
  xmlNewProp (node,BAD_CAST "P6",BAD_CAST attr);

  snprintf (attr,OKC_ARBITRARY_STRING_LENGTH,"%1.3f",params[6]);
  
  xmlNewProp (node,BAD_CAST "P7",BAD_CAST attr);

  return node;
}

/**
 * Helper function intern to OpenKC. Creates a XML-Node holding an integer 
 * parameter.
 *
 * @param node #xmlNode that will be filled with int param.
 *
 * @param param parameter to be used in XML-Node.
 *
 * @return Pointer to new #xmlNode
 */

static xmlNode*
_write_int_to_node (xmlNode* node, int param)
{
  char attr[OKC_ARBITRARY_STRING_LENGTH];

  xmlNode* text;

  snprintf (attr, OKC_ARBITRARY_STRING_LENGTH, "%d", param);

  text = xmlNewText (BAD_CAST attr);
  
  xmlAddChild (node, text);

  return node;
}



/**
 * Helper function intern to OpenKC. Creates a XML-Node for the IPOC 
 * value, which is used by the robot to know, which "Question" the XML-Answer
 * belongs to.
 *
 * @param node #xmlNode that will be filled with akorr.
 *
 * @param ipoc long long int to be copied to xmlNode.
 *
 * @return Pointer to new #xmlNode
 */



static xmlNode*
_write_ipoc_to_node (xmlNode* node, long long int ipoc)
{

  char attr[OKC_ARBITRARY_STRING_LENGTH];

  xmlNode* text;
  
  snprintf (attr, OKC_ARBITRARY_STRING_LENGTH, "%lld", ipoc);

  text = xmlNewText (BAD_CAST attr);

  xmlAddChild (node,text);
  
  return node;
}

/**
 * Helper function intern to OpenKC. Creates a XML-Node for the ESTR Node.
 * This can be used to transmit an error string to the robots KCP screen.
 *
 * @param node #xmlNode that will be filled with akorr.
 *
 * @param estr string to be send to robot.
 *
 * @return Pointer to new #xmlNode
 */

static xmlNode*
_write_estr_to_node (xmlNode* node, char* estr)
{

  xmlNode* text;
  
  text = xmlNewText (BAD_CAST estr);

  xmlAddChild (node,text);
  
  return node;
}

/**
 * Helper function intern to OpenKC. This function uses different helper 
 * functions to build up a #xmlDoc tree which then can be send to the robot.
 *
 * @param message use this error message string to the KCP.
 * @param rkorr use this #coords as position corrections.
 * @param akorr use this #lbr_axis as axis correction.
 * @param send_break use this to signal a break to the robot.
 * @param params array of 7 doubles used to signal parameters primarily for 
 *        impedance control.
 * @param command set command (normal operation, set parameter, switch 
 *        controller...
 * @param ipoc IPOC value has to be returned to robot.
 *
 * @return Pointer to new #xmlNode
 */

static xmlDoc*
_create_xml_answer (char* message, coords rkorr, lbr_axis akorr, 
		    int send_break, double* params, int command, 
		    int subcommand, int forceprofile, int forceimmediate, 
		    int forcetiming, long long int ipoc)
{
  xmlDoc* doc = NULL;
  xmlNode* root_node = NULL;
  xmlNode* node = NULL;

  char OpenKC[256];

//  printf("!!! create_xml_answer rkorr.x: |%f| \n",rkorr.x);
//  printf("!!! create_xml_answer rkorr.y: |%f| \n",rkorr.y);
//  printf("!!! create_xml_answer rkorr.z: |%f| \n",rkorr.z);

  snprintf (OpenKC,256,"OpenKC_%s",VERSION);
  
  doc = xmlNewDoc (BAD_CAST "1.0");

  if (NULL == doc)
    {
      fprintf (stderr,"_create_xml_answer: could not create xmDoc\n");
      exit (EXIT_FAILURE);
    }
  
  root_node = xmlNewNode (NULL,BAD_CAST "Sen");
  
  if (NULL == root_node)
    {
      fprintf (stderr,"_create_xml_answer: could not create xmlNode\n");
      exit (EXIT_FAILURE);
    }

  xmlNewProp(root_node, BAD_CAST "Type", BAD_CAST OpenKC);
  xmlDocSetRootElement(doc, root_node);

  if (NULL != message)
    {
      node = xmlNewChild (root_node, NULL, BAD_CAST "EStr", NULL);
      
      node = _write_estr_to_node (node, message);
    }

  node = xmlNewChild (root_node, NULL, BAD_CAST "RKorr", NULL);

  node = _write_rkorr_to_node (node,rkorr);

  node = xmlNewChild (root_node, NULL, BAD_CAST "AKorr", NULL);

  node = _write_akorr_to_node (node,akorr);

  node = xmlNewChild (root_node, NULL, BAD_CAST "EKorr", NULL);

  node = _write_ekorr_to_node (node,akorr);

  if (send_break)
    {
      node = xmlNewChild (root_node, NULL, BAD_CAST "Break", NULL);

      node = _write_int_to_node (node, send_break);
    }

  node = xmlNewChild (root_node, NULL, BAD_CAST "Param", NULL);

  node = _write_params_to_node (node, params);

  node = xmlNewChild (root_node, NULL, BAD_CAST "Command", NULL);
  
  node = _write_int_to_node (node, command);
  
  node = xmlNewChild (root_node, NULL, BAD_CAST "Subcommand", NULL);
  
  node = _write_int_to_node (node, subcommand);

  node = xmlNewChild (root_node, NULL, BAD_CAST "ForceProfile",NULL);
  
  node = _write_int_to_node (node, forceprofile);

  node = xmlNewChild (root_node, NULL, BAD_CAST "ForceImmediate", NULL);
  
  node = _write_int_to_node (node, forceimmediate);

  node = xmlNewChild (root_node, NULL, BAD_CAST "ForceTiming", NULL);
  
  node = _write_int_to_node (node, forcetiming);

  node = xmlNewChild (root_node, NULL, BAD_CAST "IPOC", NULL);

  node = _write_ipoc_to_node (node, ipoc);

  return doc;
}

/**
 * Helper function intern to OpenKC. Taken from the Socket Programming Howto
 * by Gordon McMillan. Returns in_addr regardless if its IPv4 or IPv6.
 *
 * @return pointer to in_address
 */  

static void* 
_get_in_addr(struct sockaddr *sa)
{
  if (sa->sa_family == AF_INET) 
    {
      return &(((struct sockaddr_in*)sa)->sin_addr);
    }
  
  return &(((struct sockaddr_in6*)sa)->sin6_addr);
}


/**
 * Helper function intern to OpenKC. This function registers a new robot 
 * in the okc_handle for further use. 
 *
 * @param remoteaddr Type #struct sockaddr_storage, holds network address
 *        to be registers/associated with the robot handle
 * @param okc The #okc_handle_t where the robot should be registered in.
 *
 * @return Slot of the new robot, or OKC_ERROR on error.
 */  

static int
_register_robot(struct sockaddr_storage remoteaddr, okc_handle_t *okc) 
{
  int i,j;
  if (0 != pthread_rwlock_wrlock (okc->robot_map_names_lock))
    {
      perror ("_register_robot: pthread_rwlock_wrlock");
      exit (EXIT_FAILURE);
    } 
  /* lock aquired*/
  i=0;
  while ((okc->robot_map[i] != OKC_SLOT_UNASSIGNED) &&
	 (i < OKC_MAX_ROBOTS))
    i++;
  
  if (OKC_SLOT_IN_USE == okc->robot_map[i])
    {
      if (0 != pthread_rwlock_unlock (okc->robot_map_names_lock))
	{
	  perror ("_register_robot: pthread_rwlock_unlock");
	  exit (EXIT_FAILURE);
	}
      return (OKC_ERROR);
    }
  
  okc->robot_map[i] = OKC_SLOT_IN_USE;
  okc->command[i] = OKC_NORMAL_OP;
  okc->subcommand[i] = OKC_NORMAL_OP;
  okc->send_break[i] = 0;
  okc->pending_command[i] = OKC_NORMAL_OP;
  okc->forceimmediate[i] = -1;
  okc->forceprofile[i] = -1;
  okc->forcetiming[i] = -1;
  for (j = 0; j < 7; j++)
    okc->params[i][j] = 0.0;
  
  if (NULL == okc->robot_names[i])
    {
      okc->robot_names[i] = malloc (INET6_ADDRSTRLEN*
				    sizeof (char));
      if (NULL == okc->robot_names[i])
	{
	  perror ("_register_robot: malloc");
	  exit (EXIT_FAILURE);
	}
    }

  if (okc->mode == OKC_MODE_CALLBACK_AXIS ||
      okc->mode == OKC_MODE_CALLBACK_POS)
    {
      if (0 != pthread_rwlock_rdlock (&okc->callback_lock[i]))
	{
	  perror ("_register_robot: pthread_rwlock_rdlock");
	  exit (EXIT_FAILURE);
	} 
      if (okc->update_correction_callback[i] == NULL)
	{
	  fprintf (stderr,"_register_robot: warning: no callback function "
		   "registered for robot %d\n",i);
	}
      if (0 != pthread_rwlock_unlock (&okc->callback_lock[i]))
	{
	  perror ("_register_robot: pthread_rwlock_unlock");
	  exit (EXIT_FAILURE);
	}  
    }

  if (NULL == inet_ntop (remoteaddr.ss_family,
			 _get_in_addr((struct sockaddr*)
				      &remoteaddr),
			 okc->robot_names[i], INET6_ADDRSTRLEN))
    {
      perror ("_register_robot: inet_ntop");
      strncpy (okc->robot_names[i],"unknown",INET6_ADDRSTRLEN);
    }

  if (0 != pthread_rwlock_unlock (okc->robot_map_names_lock))
    {
      perror ("_register_robot: pthread_rwlock_unlock");
      exit (EXIT_FAILURE);
    }
  
  return(i);
}

/**
 * Helper function intern to OpenKC. This function unregisters a previously 
 * registered robot in the okc_handle. 
 *
 * @param okc The #okc_handle_t where the robot should be unregistered from.
 * @param robot_id id of the robot that is to be unregistered.
 *
 * @return OKC_OK on success.
 */

static int
_unregister_robot(okc_handle_t *okc, int robot_id) 
{
  if (0 != pthread_rwlock_wrlock (okc->robot_map_names_lock))
    {
      perror ("_unregister_robot: pthread_rwlock_wrlock");
      exit (EXIT_FAILURE);
    } 
  /* lock aquired*/
  okc->robot_map[robot_id] = OKC_SLOT_UNASSIGNED;
  free (okc->robot_names[robot_id]);
  okc->robot_names[robot_id] = NULL;
  if (0 != pthread_rwlock_unlock (okc->robot_map_names_lock))
    {
      perror ("_unregister_robot: pthread_rwlock_unlock");
      exit (EXIT_FAILURE);
    }

  if (0 != pthread_rwlock_wrlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("_unregister_robot: pthread_rwlock_wrlock");
      exit (EXIT_FAILURE);
    }
  
  okc_coords_set_all_zero (&okc->pos_act[robot_id]);
  okc_coords_set_all_zero (&okc->pos_des[robot_id]);
  okc_axis_set_all_zero (&okc->axis_act[robot_id]);
  okc_axis_set_all_zero (&okc->axis_des[robot_id]);
  okc_axis_set_all_zero (&okc->torques_axis_act[robot_id]);
  okc_coords_set_all_zero (&okc->ft_tcp_est[robot_id]);
  okc->status[robot_id] = OKC_NORMAL_OP;
  
  if (0 != pthread_rwlock_unlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("_unregister_robot: pthread_rwlock_unlock");
      exit (EXIT_FAILURE);
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("_unregister_robot: pthread_rwlock_wrlock");
      exit (EXIT_FAILURE);
    }
  
  okc_coords_set_all_zero (&okc->pos_correction[robot_id]);
  okc_axis_set_all_zero (&okc->axis_correction[robot_id]);
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->subcommand[robot_id] = OKC_NORMAL_OP;
  memset (okc->params[robot_id],0,sizeof (double)*7);

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("_unregister_robot: pthread_rwlock_unlock");
      exit (EXIT_FAILURE);
    }
  
  return(OKC_OK);
}

/**
 * Helper function intern to OpenKC. Gets the intern value of the command
 * field of selected robot.
 *
 * @param okc The #okc_handle_t of the robot set.
 * @param robot_id id of the robot that is to be unregistered.
 *
 * @return value of command-field
 */

static int 
_okc_get_command (okc_handle_t* okc, int robot_id)
{
  int ret;
  if (0 != pthread_rwlock_rdlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("_okc_get_command: pthread_rwlock_rdlock");
      exit (EXIT_FAILURE);
    }

  ret = okc->command[robot_id];
  
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("_okc_get_command: pthread_rwlock_unlock");
      exit (EXIT_FAILURE);
    }
  return ret;
}



/**
 * Helper function intern to OpenKC. Takes the message from the robot and 
 * parses it and refreshes all corresponding values in the okc_handle_t.
 *
 * @param okc The #okc_handle_t of the robot set.
 * @param buffer message buffer.
 * @param nbytes length of message buffer.
 * @param robot_id id of the robot that is updated.
 *
 * @return returns ipoc value of the message.
 */

static long long int
_update_robot_okc (okc_handle_t* okc, char* buffer, int nbytes, int robot_id)
{

  xmlDoc* doc = NULL;
  long long int ipoc=0;
  coords pos_act;
  coords pos_des;
  lbr_axis axis_act;
  lbr_axis axis_des;
  lbr_axis torques_axis_act;
  coords ft_tcp_est;
  int status=0;
  int cmd;
 

  okc_coords_set_all_zero (&pos_act);
  okc_coords_set_all_zero (&pos_des);
  okc_axis_set_all_zero (&axis_act);
  okc_axis_set_all_zero (&axis_des);
  okc_axis_set_all_zero (&torques_axis_act);
  okc_coords_set_all_zero (&ft_tcp_est);


  doc = xmlReadMemory (buffer, nbytes ,"noname.xml",NULL,0);
  
  _parse_tree (doc, &pos_act, &pos_des, &axis_act, &axis_des, &torques_axis_act, 
	       &ft_tcp_est, &status,&ipoc);

  cmd = _okc_get_command (okc,robot_id);
 

  if (0 != pthread_rwlock_wrlock (&okc->rist_lbra_out_lock [robot_id]))
    {
      perror ("_update_robot_okc: pthread_rwlock_wrlock");
      exit (EXIT_FAILURE);
    }
  
  okc->pos_act[robot_id] = pos_act;
  okc->pos_des[robot_id] = pos_des;
  okc->axis_act[robot_id] = axis_act;
  okc->axis_des[robot_id] = axis_des;
  okc->torques_axis_act[robot_id] = torques_axis_act;
  okc->ft_tcp_est[robot_id] = ft_tcp_est;
  okc->status[robot_id] = status;
  okc->update_ticks[robot_id]++;
  if (cmd == OKC_NORMAL_OP)
    {
      okc->pending_command[robot_id] = OKC_NORMAL_OP;
      okc->ticks_command[robot_id] = 0;
    }
  else
    {
      if (okc->ticks_command[robot_id] == 0)
	{
	  okc->pending_command[robot_id] = okc->command[robot_id];
	  okc->command[robot_id] = OKC_COMMAND_PENDING;
	  okc->ticks_command[robot_id] = 1;
	}
      else
	{
	  okc->ticks_command[robot_id]++;
	}
    }

  if (0 != pthread_rwlock_unlock (&okc->rist_lbra_out_lock [robot_id]))
    {
      perror ("_update_robot_okc: pthread_rwlock_wrlock");
      exit (EXIT_FAILURE);
    }

  xmlFreeDoc (doc);
	
  return ipoc;
}


/**
 * Helper function intern to OpenKC. Creates an XML-answer for the robot, using
 * the variables set in okc_handle_t
 *
 * @param okc The #okc_handle_t of the robot set.
 * @param buffer message buffer.
 * @param nbytes length of message buffer.
 * @param robot_id id of the robot that is updated.
 *
 * @return returns ipoc value of the message.
 */

static int
_create_answer_for_robot (okc_handle_t* okc, xmlChar** outbuf, int* len_outbuf, 
			  int robot_id, long long int ipoc)
{
  xmlDoc* doc = NULL;
  
  coords k;
  lbr_axis l;
  double params[7];
  int command;
  int subcommand;
  int send_break;
  int forceprofile;
  int forceimmediate;
  int forcetiming;

  if (0 != pthread_rwlock_rdlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("_create_answer_for_robot: pthread_rwlock_rdlock");
      exit (EXIT_FAILURE);
    }

  k = okc->pos_correction[robot_id];
  l = okc->axis_correction[robot_id];
  memcpy (params,okc->params[robot_id],sizeof(double)*7);
  send_break = okc->send_break[robot_id];
  subcommand = okc->subcommand[robot_id];
  forceprofile = okc->forceprofile[robot_id];
  forceimmediate = okc->forceimmediate[robot_id];
  forcetiming = okc->forcetiming[robot_id];

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("_create_answer_for_robot: pthread_rwlock_unlock");
      exit (EXIT_FAILURE);
    }

  if (0 != pthread_rwlock_rdlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("_create_answer_for_robot: pthread_rwlock_rdlock");
      exit (EXIT_FAILURE);
    }
  if (1 == (okc->ticks_command[robot_id] % 130))
    command = okc->pending_command[robot_id];
  else
    command = okc->command[robot_id];


  if (0 != pthread_rwlock_unlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("_create_answer_for_robot: pthread_rwlock_unlock");
      exit (EXIT_FAILURE);
    }
  
  doc = _create_xml_answer (NULL, k, l, send_break,params,command,
			    subcommand, forceprofile,
			    forceimmediate,
			    forcetiming,ipoc);

  xmlDocDumpMemory(doc, outbuf, len_outbuf);

		      
  xmlFreeDoc (doc);
  
  return OKC_OK;
}

/**
 * Helper function intern to OpenKC. Locks rwlock and sets the new cartesian
 * corrections to be send.
 *
 * @param okc The #okc_handle_t of the robot set.
 * @param robot_id Number of robot
 * @param k_correction The #coords to be set.
 *
 */
static void
_okc_set_coords_correction (okc_handle_t* okc, int robot_id, coords k_correction)
{

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("_okc_set_coords_correction: pthread_rwlock_wrlock");
      exit (EXIT_FAILURE);
    }
  
  okc->pos_correction[robot_id] = k_correction;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("_okc_set_coords_correction: pthread_rwlock_wrlock");
      exit (EXIT_FAILURE);
    }
}
/**
 * Helper function intern to OpenKC. Locks rwlock and sets the new axis
 * corrections to be send.
 *
 * @param okc The #okc_handle_t of the robot set.
 * @param robot_id Number of robot
 * @param a_correction The #lbr_axis to be set.
 *
 */
static void
_okc_set_axis_correction (okc_handle_t* okc, int robot_id, 
			     lbr_axis a_correction)
{
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("_okc_set_axis_correction: pthread_rwlock_wrlock");
      exit (EXIT_FAILURE);
    }
  
  okc->axis_correction[robot_id] = a_correction;
  
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("_okc_set_axis_correction: pthread_rwlock_unlock");
      exit (EXIT_FAILURE);
    }
}

/**
 * Helper function intern to OpenKC. This function runs in its own thread.
 * This is the implementation of the actual server. After some setup-stuff
 * the server will wait for packages from robot(s) that need to be answered.
 * Depending on the mode, callback-functions are called from within this 
 * routine or the values set by okc_set_coords_correction or 
 * okc_set_axis_correction are used. 
 *
 * @param okc_struct_p Handle to the original #okc_handle_t. 
 *
 * @see okc_start_server
 * @see okc_stop_server
 * @see okc_set_coords_correction
 * @see okc_set_axis_correction
 */

static void*
_serve_connections (void* okc_struct_p)
{

  coords kcorr,kpos;
  lbr_axis acorr,apos;
  fd_set master;
  fd_set read_fds;
  int fdmax;
  int newfd=0;
  char remoteIP[INET6_ADDRSTRLEN];
  socklen_t addrlen;
  struct sockaddr_storage remoteaddr; 

  int nbytes,cur_fd, newrobotid=0;
  char* buffer;
  xmlChar* outbuf;
  int len_outbuf;
  short* map_fd_to_robot;
  long long int ipoc;
  struct timeval sleep_val;
#ifdef BENCH
  long long int serve=0;
  int last_cmd=0;
  int ticks=0;
  int last_status=0;
  int st_ticks=0;
  struct timeval tv_last, tv_now;
#endif
  sigset_t allsigs; /**< Ignore all signals (esp. SIGRT). */

  okc_handle_t* okc = (okc_handle_t*) okc_struct_p;

  /** Clear signal mask: This thread won't get any signals delivered. */
  sigfillset (&allsigs);
  pthread_sigmask (SIG_BLOCK, &allsigs, NULL);

  okc_coords_set_all_zero (&kcorr);
  okc_axis_set_all_zero (&acorr);

  buffer = malloc (OKC_BUFFER_SIZE*sizeof (char));

  if (NULL == buffer)
    {
      perror ("_serve_connections: malloc");
      exit (EXIT_FAILURE);
    }


  FD_ZERO(&master);
  FD_ZERO(&read_fds);

  fdmax = okc->listener; 

  FD_SET (fdmax,&master);
  
  map_fd_to_robot = malloc ((fdmax+1)*sizeof (short)); /* waste some RAM, gain
							readable code */
  if (NULL == map_fd_to_robot)
    {
      perror ("_serve_connections: malloc");
      exit (EXIT_FAILURE);
    }
#ifdef BENCH
  gettimeofday (&tv_last, NULL);
#endif
  while ((okc->flags&OKC_QUIT) != OKC_QUIT)
    {
      read_fds = master; 
      sleep_val.tv_sec = 1;
      sleep_val.tv_usec = 0;
      if (select(fdmax+1, &read_fds, NULL, NULL, &sleep_val) == -1) 
	{
	  perror("_serve_connections: select");
	  exit(EXIT_FAILURE);
	}
      
#ifdef BENCH
      gettimeofday (&tv_now,NULL);
      if (tv_now.tv_sec > tv_last.tv_sec)
	{
	  printf ("Served %lu requests last second\n",serve);
	  printf ("DEBUG: okc->flags = %d\n",okc->flags);
	  tv_last = tv_now;
	  serve = 0;
	}
#endif
      /** run through the existing connections looking for data to read */
      for(cur_fd = OKC_MIN_FD; cur_fd <= fdmax; cur_fd++) 
	{
	  if (FD_ISSET(cur_fd, &read_fds)) 
	    { /*  we got one!! */
	      if (cur_fd == okc->listener) 
		{
		  /** handle new connections */
		  addrlen = sizeof (remoteaddr);
		  newfd = accept(okc->listener,
				 (struct sockaddr *)&remoteaddr,
				 &addrlen);

		  if (newfd == -1) 
		    {
		      perror("_serve_connections: accept");
		    } 
		  else 
		    {
		      /** register new robot */
		      newrobotid = _register_robot(remoteaddr, okc);
		      if (OKC_ERROR == newrobotid)
			{
			  fprintf (stderr,"_serve_connections: warning: "
				   "could not register robot from IP %s\n"
				   "Reason: no more open slots available.\n",
				   inet_ntop(remoteaddr.ss_family,
					     _get_in_addr((struct sockaddr*)
							  &remoteaddr),
					     remoteIP, INET6_ADDRSTRLEN));
			  close (newfd);
			}
		      else
			{
			  FD_SET(newfd, &master); /** add to master set */
			  if (newfd > fdmax) 
			    {    /* keep track of the max */
			      fdmax = newfd;
			      map_fd_to_robot = realloc (map_fd_to_robot, 
							 (fdmax+1)*
							 sizeof (short));
			      if (NULL == map_fd_to_robot)
				{
				  perror ("_serve_connections: realloc");
				  exit (EXIT_FAILURE);
				}
			    }
			  map_fd_to_robot[newfd] = newrobotid;
			  printf("_serve_connections: new connection from %s on "
				 "socket %d assigned to robot id %d\n",
				 inet_ntop(remoteaddr.ss_family,
					   _get_in_addr((struct sockaddr*)
							&remoteaddr),
					   remoteIP, INET6_ADDRSTRLEN),
				 newfd, newrobotid);
			  
			}
		    }
		}
	      else 
		{
		  memset (buffer,0,OKC_BUFFER_SIZE);
		  /** handle data from a client */
		  if ((nbytes = recv(cur_fd, buffer, OKC_BUFFER_SIZE, 0)) <= 0) 
		    {
		      /** got error or connection closed by client */
		      if (nbytes == 0) 
			{
			  /** connection closed */
			  printf("_serve_connections: socket %d hung up\n", cur_fd);
			  _unregister_robot (okc, map_fd_to_robot[cur_fd]);
			} 
		      else 
			{
			  perror("_serve_connections: recv");
			}
		      close(cur_fd); /* bye! */
		      FD_CLR(cur_fd, &master); /* remove from master set */
		      map_fd_to_robot[cur_fd] = OKC_SLOT_UNASSIGNED;
		    } 
		  else 
		    {
		      /** we got some data from a client */
#ifdef DEBUG_CONNECTION		      
		      printf ("received %d bytes\n",nbytes);
		      
		      printf ("****** Start Buffer *******\n%s\n"
		              "******  End Buffer  *******\n",
		      			      buffer);
#endif
		      ipoc = _update_robot_okc (okc,buffer,nbytes,
						map_fd_to_robot[cur_fd]);
		      if (okc->mode == OKC_MODE_CALLBACK_AXIS)
			{
			  okc_get_axis_act (okc,map_fd_to_robot[cur_fd],
						&apos);
			  if (0 != pthread_rwlock_rdlock 
			      (&okc->callback_lock[map_fd_to_robot[cur_fd]]))
			    {
			      perror ("_serve_connections: "
				      "pthread_rwlock_rdlock");
			      exit (EXIT_FAILURE);
			    } 
			  if (okc->update_correction_callback[map_fd_to_robot
							      [cur_fd]] != NULL)
			    {
			      if (OKC_OK == 
				  okc->update_correction_callback[map_fd_to_robot
								  [cur_fd]] 
				  (okc->private_to_update_correction_callback[map_fd_to_robot
								  [cur_fd]], 
				   (double*) &apos,
				   (double*) &acorr))
				_okc_set_axis_correction (okc,
							  map_fd_to_robot
							  [cur_fd],
							  acorr);
			      _okc_set_coords_correction (okc,
			      							    map_fd_to_robot
			      							    [cur_fd],
			      							    kcorr);
			    }
			  if (0 != pthread_rwlock_unlock 
			      (&okc->callback_lock[map_fd_to_robot[cur_fd]]))
			    {
			      perror ("_register_robot: pthread_rwlock_unlock");
			      exit (EXIT_FAILURE);
			    } 
			}
		      if (okc->mode == OKC_MODE_CALLBACK_POS)
			{
			  if (0 != pthread_rwlock_rdlock 
			      (&okc->callback_lock[map_fd_to_robot[cur_fd]]))
			    {
			      perror ("_serve_connections: "
				      "pthread_rwlock_rdlock");
			      exit (EXIT_FAILURE);
			    } 
			  okc_get_coords_act (okc,map_fd_to_robot[cur_fd],
						  &kpos);
			  if (okc->update_correction_callback[map_fd_to_robot
							      [cur_fd]] != NULL)
			    {
			      if (OKC_OK == 
				  okc->update_correction_callback[map_fd_to_robot
								  [cur_fd]] 
				  (okc->private_to_update_correction_callback[map_fd_to_robot
								  [cur_fd]], 
				   (double*) &kpos,
				   (double*) &kcorr))
				_okc_set_coords_correction (okc,
							    map_fd_to_robot
							    [cur_fd],
							    kcorr);
			    }
			  if (0 != pthread_rwlock_unlock 
			      (&okc->callback_lock[map_fd_to_robot[cur_fd]]))
			    {
			      perror ("_register_robot: pthread_rwlock_unlock");
			      exit (EXIT_FAILURE);
			    } 
			}
			      
								       
		      
		      _create_answer_for_robot (okc, &outbuf, &len_outbuf,
						map_fd_to_robot[cur_fd],
						ipoc);


			      
		      /* if ((okc->command[map_fd_to_robot[cur_fd]]) || */
/* 			  (okc->status[map_fd_to_robot[cur_fd]]) ) */
/* 			printf ("Got: %s\n Answer: ***\n%s\n",buffer,outbuf); */
#ifdef DEBUG_CONNECTION
		      printf ("~~~~~Sending~~~~~\n%s-------------------\n",
			      outbuf);
#endif
		      send (cur_fd, outbuf, len_outbuf, 0);
		      

		      
		      
		      xmlFree (outbuf);


#ifdef BENCH
		      if (okc->command[map_fd_to_robot[cur_fd]] == last_cmd)
			ticks++;
		      else
			{
			  fprintf (stderr,"Command %d lastet %d ticks\n",
				   last_cmd,ticks);
			  last_cmd = okc->command[map_fd_to_robot[cur_fd]];
			  ticks = 0;
			}
		      if (okc->status[map_fd_to_robot[cur_fd]] == last_status)
			st_ticks++;
		      else
			{
			  fprintf (stderr,"Status %d lastet %d ticks\n",
				   last_status,st_ticks);
			  last_status = okc->status[map_fd_to_robot[cur_fd]];
			  st_ticks = 0;
			}
		      serve++;
#endif
		  
		    }
		} 
	    }
	}
    }

  free (buffer); 
  
  return (NULL);
}

/**
 * Checks, if a certain robot_id is associated to a robot. Returns #OKC_OK
 * if robot_id is connected to a real robot, or #OKC_UNAVAILABLE if no 
 * physical robot has connected to robot_id.
 *
 * @param okc Pointer to #okc_handle_t. 
 * @param robot_id Robot ID to check.
 * @return OKC_OK, if robot is connected, OKC_UNAVAILABLE if no robot is
 *         currently connected or OKC_ERROR on error.
 *
 */

int
okc_is_robot_avail (okc_handle_t* okc, int robot_id)
{

  int i;

  if (robot_id >= OKC_MAX_ROBOTS)
    {
      fprintf (stderr,"okc_is_robot_avail: invalid robot_id %d. Max is %d\n",
	       robot_id, OKC_MAX_ROBOTS - 1);
      return OKC_ERROR;
    }

  if (0 != pthread_rwlock_rdlock (okc->robot_map_names_lock))
    {
      perror ("okc_is_robot_avail: pthread_rwlock_rdlock");
      return (OKC_ERROR);
    }

  if (OKC_SLOT_IN_USE == okc->robot_map[robot_id])
    i = OKC_OK;
  else
    i = OKC_UNAVAILABLE;

  if (0 != pthread_rwlock_unlock (okc->robot_map_names_lock))
    {
      perror ("okc_is_robot_avail: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return i;
}


/**
 * With this function, which is only available if OKC_MODE_SETTER was
 * used when creaing the okc server, the cartesian trajectorie
 * correction can be set. The correction is valid and will be
 * communicated to the robot as long as it is set. Currently, there
 * is no way to tell that the correction did take place. One has to
 * contol the robots position through #okc_get_coords_act and possibly
 * adapt to the robots IPO-Cycle by, for example, a timer, or use the
 * OKC_MODE_CALLBACK_* modes.
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param k_correction Cartesian corrections.
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 *
 */
int
okc_set_coords_correction (okc_handle_t* okc, int robot_id, coords k_correction)
{
  lbr_axis l;

  okc_axis_set_all_zero (&l);
  
  if (okc->mode != OKC_MODE_SETTER)
    {
      fprintf (stderr,"okc_set_coords_correction: function only available in"
	       " OKC_MODE_SETTER-mode!\n");
      return (OKC_ERROR);
    }

  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_set_coords_correction: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_coords_correction: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->pos_correction[robot_id] = k_correction;
  okc->axis_correction[robot_id] = l;
  
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_coords_correction: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  return (OKC_OK);
}


/**
 * With this function, which is only available if OKC_MODE_SETTER was
 * used when creaing the okc server, the axis trajectorie correction
 * can be set. The correction is valid and will be communicated to the
 * robot as long as it is set. Currently, there is no way to tell
 * that the correction did take place. One has to contol the robots
 * position through #okc_get_axis_act and possibly adapt to the robots
 * IPO-Cycle by, for example, a timer, or use the OKC_MODE_CALLBACK_*
 * modes.
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param a_correction Axis corrections.
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 *
 */

int
okc_set_axis_correction (okc_handle_t* okc, int robot_id, 
			     lbr_axis a_correction)
{
  
  coords k;

  okc_coords_set_all_zero (&k);

  if (okc->mode != OKC_MODE_SETTER)
    {
      fprintf (stderr,"okc_set_axis_correction: function only available in"
	       " OKC_MODE_SETTER-mode!\n");
      return (OKC_ERROR);
    }

  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_set_axis_correction: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }



  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_axis_correction: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->pos_correction[robot_id] = k;
  okc->axis_correction[robot_id] = a_correction;
  
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_axis_correction: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  
  return (OKC_OK);
}

/**
 * Get the robots status flag. The status is used by OpenKukaControl
 * to acknowledge certain mode-changes or to handshake data transmits.
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param status Axis corrections.
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 *
 */

int
okc_get_status (okc_handle_t* okc, int robot_id, int* status)
{
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_get_status: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }
  
  if (0 != pthread_rwlock_rdlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_status: pthread_rwlock_rdlock");
      return (OKC_ERROR);
    }
  
  (*status) = okc->status[robot_id];
  
  if (0 != pthread_rwlock_unlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_status: pthread_rwlock_unlock");
      return (OKC_ERROR);
    } 
  return (OKC_OK);
}

/**
 * Sets the axis stiffness parameters. Altough the parameters are
 * transmitted to the robot, the (new) parameters are only copied to
 * local variables in the KRL-program. To actually use and update the
 * parameters on the robots, you have to "switch" to a controller,
 * even if you are already using this controller. Do not forget to set
 * the dampings also to a reasonable value.
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param a_stiff Stiffness for each axis. Consult the KUKA LBR manual for 
 *        details. Range is 0.01 -- 2000.0, Default 1000.0
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 *
 * @see okc_switch_to_cp_impedance
 * @see okc_switch_to_axis_impedance
 */

int
okc_set_axis_stiffness (okc_handle_t* okc, int robot_id, 
			    lbr_axis a_stiff)
{
  struct timespec ts;
  int status=0;

  int i;

  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_set_axis_stiffness: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  for (i=0; i < 100; i++)
    {
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,
		   "okc_set_axis_stiffness: unable to retrieve status\n");
	  return (OKC_ERROR);
	}
      if (OKC_OK == status)
	break;
      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_axis_stiffness: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  /* Could do this much easier, but this will make us independant from 
     axis odering in struct lbr_axis */
  okc->params[robot_id][0] = a_stiff.a1;
  okc->params[robot_id][1] = a_stiff.a2;
  okc->params[robot_id][2] = a_stiff.a3;
  okc->params[robot_id][3] = a_stiff.a4;
  okc->params[robot_id][4] = a_stiff.a5;
  okc->params[robot_id][5] = a_stiff.a6;
  okc->params[robot_id][6] = a_stiff.e1;

  okc->command[robot_id] = OKC_SET_AXIS_STIFFNESS;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_axis_stiffness: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  while (status != OKC_ACK_SET_AXIS_STIFFNESS)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_set_axis_stiffness: could not retrieve robots"
		   " status\n");
	  return (OKC_ERROR);
	}
    }
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_axis_stiffness: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_axis_stiffness: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}


/**
 * Sets the axis damping parameters. Altough the parameters are
 * transmitted to the robot, the (new) parameters are only copied to
 * local variables in the KRL-program. To actually use and update the
 * parameters on the robots, you have to "switch" to a controller,
 * even if you are already using this controller. Do not forget to set
 * the stiffness also to a reasonable value.
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param a_damp Dampings for each axis. Consult the KUKA LBR manual for 
 *        details. Range is 0.1 -- 1.0, Default 0.7
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 *
 * @see okc_switch_to_cp_impedance
 * @see okc_switch_to_axis_impedance
 */


int
okc_set_axis_damping (okc_handle_t* okc, int robot_id, 
			    lbr_axis a_damp)
{
  struct timespec ts;
  int status=0;
  int i;
  
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_set_axis_damping: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }

    for (i=0; i < 100; i++)
    {
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,
		   "okc_set_axis_damping: unable to retrieve status\n");
	  return (OKC_ERROR);
	}
      if (OKC_OK == status)
	break;
      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_axis_damping: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  /* Could do this much easier, but this will make us independant from 
     axis odering in struct lbr_axis */
  okc->params[robot_id][0] = a_damp.a1;
  okc->params[robot_id][1] = a_damp.a2;
  okc->params[robot_id][2] = a_damp.a3;
  okc->params[robot_id][3] = a_damp.a4;
  okc->params[robot_id][4] = a_damp.a5;
  okc->params[robot_id][5] = a_damp.a6;
  okc->params[robot_id][6] = a_damp.e1;

  okc->command[robot_id] = OKC_SET_AXIS_DAMPING;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_axis_damping: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  while (status != OKC_ACK_SET_AXIS_DAMPING)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_set_axis_damping: could not retrieve robots"
		   " status\n");
	  return (OKC_ERROR);
	}
    }
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_axis_damping: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_axis_damping: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}


/**
 * Sets the cartesian stiffness parameters. Altough the parameters are
 * transmitted to the robot, the (new) parameters are only copied to
 * local variables in the KRL-program. To actually use and update the
 * parameters on the robots, you have to "switch" to a controller,
 * even if you are already using this controller. Do not forget to set
 * the dampings also to a reasonable value. (Also set AXIS stiffness
 * and damping for nullspace control.)
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param k_stiff Stiffness for each component. Consult the KUKA LBR manual for 
 *        details. Range is 0.01 -- 5000.0, Default 2000.0 for X,Y,Z and
 *        0.01 -- 300 (Default 200) for A,B and C.
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 *
 * @see okc_switch_to_cp_impedance
 */



int
okc_set_cp_stiffness (okc_handle_t* okc, int robot_id, 
			  coords k_stiff)
{
  struct timespec ts;
  int status=0;
  int i;
  
  
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_set_cp_stiffness: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }
  
  for (i=0; i < 100; i++)
    {
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,
		   "okc_set_cp_stiffness: unable to retrieve status\n");
	  return (OKC_ERROR);
	}
      if (OKC_OK == status)
	break;
      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_cp_stiffness: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->params[robot_id][0] = k_stiff.x;
  okc->params[robot_id][1] = k_stiff.y;
  okc->params[robot_id][2] = k_stiff.z;
  okc->params[robot_id][3] = k_stiff.a;
  okc->params[robot_id][4] = k_stiff.b;
  okc->params[robot_id][5] = k_stiff.c;
  okc->params[robot_id][6] = 0.0;

  okc->command[robot_id] = OKC_SET_CP_STIFFNESS;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_cp_stiffness: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  while (status != OKC_ACK_SET_CP_STIFFNESS)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_set_cp_stiffness: could not retrieve robots"
		   " status\n");
	  return (OKC_ERROR);
	}
    }
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_cp_stiffness: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_cp_stiffness: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}


/**
 * Sets the cartesian damping parameters. Altough the parameters are
 * transmitted to the robot, the (new) parameters are only copied to
 * local variables in the KRL-program. To actually use and update the
 * parameters on the robots, you have to "switch" to a controller,
 * even if you are already using this controller. Do not forget to set
 * the stiffness also to a reasonable value. (Also set AXIS stiffness
 * and damping for nullspace control.)
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param k_stiff Stiffness for each component. Consult the KUKA LBR manual for 
 *        details. Range is 0.01 -- 5000.0, Default 2000.0 for X,Y,Z and
 *        0.01 -- 300 (Default 200) for A,B and C.
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 *
 * @see okc_switch_to_cp_impedance
 */

int
okc_set_cp_damping (okc_handle_t* okc, int robot_id, 
			coords k_damp)
{
  struct timespec ts;
  int status=0;
  int i;

  
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_set_cp_damping: invalid/unavailable robot_id\n");
      return (OKC_ERROR);
    }

  for (i=0; i < 100; i++)
    {
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,
		   "okc_set_cp_damping: unable to retrieve status\n");
	  return (OKC_ERROR);
	}
      if (OKC_OK == status)
	break;
      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_cp_damping: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->params[robot_id][0] = k_damp.x;
  okc->params[robot_id][1] = k_damp.y;
  okc->params[robot_id][2] = k_damp.z;
  okc->params[robot_id][3] = k_damp.a;
  okc->params[robot_id][4] = k_damp.b;
  okc->params[robot_id][5] = k_damp.c;
  okc->params[robot_id][6] = 0.0;

  okc->command[robot_id] = OKC_SET_CP_DAMPING;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_cp_damping: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  while (status != OKC_ACK_SET_CP_DAMPING)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_set_cp_damping: could not retrieve robots"
		   " status\n");
	  return (OKC_ERROR);
	}
    }
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_cp_damping: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif  
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;
  
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_cp_damping: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}


/**
 * Switch to cartesian impedance control. The switch will take some
 * time, during which all corrections should be set to 0.0, otherwise,
 * the Kuka Robot might bail out. When this function returns, robot is
 * in cartesian impedance control (check Kuka manual on this). With
 * switch to this controller, previous set values in stiffness and
 * damping will be activated (axis stiffness/damping for nullspace control).
 *

 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * 
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 *
 * @see okc_set_axis_stiffness
 * @see okc_set_axis_damping
 * @see okc_set_cp_stiffness
 * @see okc_set_cp_damping
 * @see okc_switch_to_gravcomp
 * @see okc_switch_to_axis_impedance
 * @see okc_switch_to_position
 */


int
okc_switch_to_cp_impedance (okc_handle_t* okc, int robot_id)
{
  struct timespec ts;
  int status=0;
  int i;
  
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_switch_to_cp_impedance: invalid/unavailable "
	       "robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  for (i=0; i < 100; i++)
    {
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,
		   "okc_switch_to_cp_impedance: unable to retrieve status\n");
	  return (OKC_ERROR);
	}
      if (OKC_OK == status)
	break;
      
      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_cp_impedance: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->command[robot_id] = OKC_SET_SWITCH_CP_IMPEDANCE;
  okc->send_break[robot_id] = 1;
    
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_cp_impedance: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  while (status != OKC_ACK_SET_SWITCH_CP_IMPEDANCE)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_switch_to_cp_impedance: could not "
		   "retrieve robots status\n");
	  return (OKC_ERROR);
	}
    }
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_cp_impedance: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif  
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->send_break[robot_id] = 0;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;
  
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_cp_impedance: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}

/**
 * Switch to axis impedance control. The switch will take some time,
 * during which all corrections should be set to 0.0, otherwise, the
 * Kuka Robot might bail out. When this function returns, robot is in
 * axis impedance control (check Kuka manual on this). With switch to
 * this controller, previous set values in axis stiffness and damping
 * will be activated.
 *

 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * 
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 *
 * @see okc_set_axis_stiffness
 * @see okc_set_axis_damping
 * @see okc_switch_to_cp_impedance
 * @see okc_switch_to_gravcomp
 * @see okc_switch_to_position
 */


int
okc_switch_to_axis_impedance (okc_handle_t* okc, int robot_id)
{
  struct timespec ts;
  int status=0;
  int i;
  
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_switch_to_axis_impedance: invalid/unavailable "
	       "robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  for (i=0; i < 100; i++)
    {
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,
		   "okc_switch_to_axis_impedance: unable to retrieve status\n");
	  return (OKC_ERROR);
	}
      if (OKC_OK == status)
	break;
      
      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_axis_impedance: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->command[robot_id] = OKC_SET_SWITCH_AXIS_IMPEDANCE;
  okc->send_break[robot_id] = 1;


  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_axis_impedance: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  while (status != OKC_ACK_SET_SWITCH_AXIS_IMPEDANCE)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_switch_to_axis_impedance: could not "
		   "retrieve robots status\n");
	  return (OKC_ERROR);
	}
    }
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_axis_impedance: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif  
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->send_break[robot_id] = 0;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;
  
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_axis_impedance: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}

/**
 * Switch to gravitation compensation mode. The switch will take a
 * short time. After this function returns, robot is in GravComp mode
 * (check Kuka manual on this). You should have taken care that you
 * have choosen the correct tool and that $GRAVITATION in the KRL
 * environment is set correctly. Bad things might happen otherwise.
 *
 * @warning Check Tool and Gravitation settings on the controller!!!

 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * 
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 * @see okc_switch_to_cp_impedance
 * @see okc_switch_to_axis_impedance
 * @see okc_switch_to_position
 */


int
okc_switch_to_gravcomp (okc_handle_t* okc, int robot_id)
{
  struct timespec ts;
  int status=0;
  int i;

  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_switch_to_gravcomp: invalid/unavailable "
	       "robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  for (i=0; i < 100; i++)
    {
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,
		   "okc_switch_to_gravcomp: unable to retrieve status\n");
	  return (OKC_ERROR);
	}
      if (OKC_OK == status)
	break;
      
      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_gravcomp: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->command[robot_id] = OKC_SET_SWITCH_GRAVCOMP;
  okc->send_break[robot_id] = 1;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_gravcomp: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  while (status != OKC_ACK_SET_SWITCH_GRAVCOMP)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_switch_gravcomp: could not "
		   "retrieve robots status\n");
	  return (OKC_ERROR);
	}
    }
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_gravcomp: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif  
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->send_break[robot_id] = 0;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;
  
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_gravcomp: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}


/**
 * Switch to position control mode. The switch will take a short time
 * during which all corrections should be set to 0.0, otherwise, the
 * Kuka Robot might bail out.. After this function returns, robot is
 * in position contol mode (check Kuka manual on this). 
 *
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * 
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 * @see okc_switch_to_cp_impedance
 * @see okc_switch_to_axis_impedance
 * @see okc_switch_to_gravcomp
 */


int
okc_switch_to_position (okc_handle_t* okc, int robot_id)
{
  struct timespec ts;
  int status=0;
  int i;
  
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_switch_to_position: invalid/unavailable "
	       "robot_id\n");
      return (OKC_UNAVAILABLE);
    }
  
  for (i=0; i < 100; i++)
    {
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,
		   "okc_switch_to_position: unable to retrieve status\n");
	  return (OKC_ERROR);
	}
      if (OKC_OK == status)
	break;
      
      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_position: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->command[robot_id] = OKC_SET_SWITCH_POSITION;
  okc->send_break[robot_id] = 1;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_position: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  while (status != OKC_ACK_SET_SWITCH_POSITION)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_switch_to_position: could not "
		   "retrieve robots status\n");
	  return (OKC_ERROR);
	}
    }
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_position: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif  
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->send_break[robot_id] = 0;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;
  
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_switch_to_position: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}

/**
 * Tells the server, to send a break signal to the Robot. The
 * break-signal will stop the ST_MOVESENS directive. This directive is
 * mostly useful from inside the OpenKC Library, but might be of some
 * use if you use a custom KRL-program.
 *
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * 
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 */


int
okc_send_break_to_robot (okc_handle_t* okc, int robot_id)
{
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_send_break_to_robot: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  if (0 != pthread_rwlock_wrlock (okc->robot_map_names_lock))
    {
      perror ("okc_send_break_to_robot: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->send_break[robot_id] = 1;
  
  if (0 != pthread_rwlock_unlock (okc->robot_map_names_lock))
    {
      perror ("okc_send_break_to_robot: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  
  return (OKC_OK);
}

/**
 * Gets the last (actual) reported axis positions from the robot. 
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param l pointer to #lbr_axis where the actual values will be copied.
 * 
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 */

int
okc_get_axis_act (okc_handle_t* okc, int robot_id, lbr_axis* l)
{
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_get_axis_act: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  if (0 != pthread_rwlock_rdlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_axis_act: pthread_rwlock_rdlock");
      return (OKC_ERROR);
    }
  
  okc_copy_axis (okc->axis_act[robot_id],l);
  
  if (0 != pthread_rwlock_unlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_axis_act: pthread_rwlock_unlock");
      return (OKC_ERROR);
    } 
  return (OKC_OK);
}

/**
 * Gets the last reported *desired* or set axis positions from the 
 * robot. Depending on the control mode, these values can be significantly 
 * different from the actual joint values.
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param l pointer to #lbr_axis where the desired values will be copied.
 * 
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 */

int
okc_get_axis_des (okc_handle_t* okc, int robot_id, lbr_axis* l)
{
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_get_axis_des: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  if (0 != pthread_rwlock_rdlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_axis_des: pthread_rwlock_rdlock");
      return (OKC_ERROR);
    }
  
  okc_copy_axis (okc->axis_des[robot_id],l);
  
  if (0 != pthread_rwlock_unlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_axis_des: pthread_rwlock_unlock");
      return (OKC_ERROR);
    } 
  return (OKC_OK);
}


/**
 * Gets the last (actual) reported axis torques from the robot. 
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param l pointer to #lbr_axis where the actual values will be copied.
 * 
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 */


int
okc_get_torques_axis_act (okc_handle_t* okc, int robot_id, lbr_axis* l)
{
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_get_torques_axis_act: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  if (0 != pthread_rwlock_rdlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_torques_axis_act: pthread_rwlock_rdlock");
      return (OKC_ERROR);
    }
  
  okc_copy_axis (okc->torques_axis_act[robot_id],l);
  
  if (0 != pthread_rwlock_unlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_torques_axis_act: pthread_rwlock_unlock");
      return (OKC_ERROR);
    } 
  return (OKC_OK);
}

/**
 * Gets the last (actual) reported cartesian position of the
 * endeffector (+ tool) from the robot.
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param k pointer to #coords where the actual values will be copied.
 * 
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 */


int
okc_get_coords_act (okc_handle_t* okc, int robot_id, coords* k)
{
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_get_coords_act: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  if (0 != pthread_rwlock_rdlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_coords_act: pthread_rwlock_rdlock");
      return (OKC_ERROR);
    }
  
  okc_copy_coords (okc->pos_act[robot_id],k);
  
  if (0 != pthread_rwlock_unlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_coords_act: pthread_rwlock_unlock");
      return (OKC_ERROR);
    } 
  return (OKC_OK);
}

/**
 * Gets the last *desired* or set cartesian position of the
 * endeffector (+ tool) from the robot. Depending on the control mode,
 * these values can be significantly different from the actual positions.
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param k pointer to #coords where the actual values will be copied.
 * 
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 */


int
okc_get_coords_des (okc_handle_t* okc, int robot_id, coords* k)
{
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_get_coords_des: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  if (0 != pthread_rwlock_rdlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_coords_des: pthread_rwlock_rdlock");
      return (OKC_ERROR);
    }
  
  okc_copy_coords (okc->pos_des[robot_id],k);
  
  if (0 != pthread_rwlock_unlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_coords_des: pthread_rwlock_unlock");
      return (OKC_ERROR);
    } 
  return (OKC_OK);
}

/**
 * Gets the last (actual) reported (cartesian) torques from the
 * endeffector. These torques are estimates only from the axis torque
 * sensor readings reported by the KRC. In some robot
 * configurations/positions, these estimates may not be anywhere
 * reliable. So be cautious. Usually, one should have a close look at
 * the variance of the estimated values, but due to limitations in
 * RSI-XML we cannot provide the internal variances from the KRC to
 * the PC.
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * @param k pointer to #coords where the actual values will be copied.
 * 
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 */

int
okc_get_ft_tcp_est (okc_handle_t* okc, int robot_id, coords* k)
{
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_get_ft_tcp_est: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  if (0 != pthread_rwlock_rdlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_ft_tcp_est: pthread_rwlock_rdlock");
      return (OKC_ERROR);
    }
  
  okc_copy_coords (okc->ft_tcp_est[robot_id],k);
  
  if (0 != pthread_rwlock_unlock (&okc->rist_lbra_out_lock[robot_id]))
    {
      perror ("okc_get_ft_tcp_est: pthread_rwlock_unlock");
      return (OKC_ERROR);
    } 
  return (OKC_OK);
}


/**
 * Function internal to OpenKC. Sets the command to "exit gracefully",
 * so the KRL programm will terminate in a nice way ;-)
 *
 * @param okc Pointer to #okc_handle_t.  
 * @param robot_id Robot ID to use.  
 * 
 * @return OKC_OK, or OKC_ERROR on error.
 */

static int
_okc_send_exit_gracefully (okc_handle_t* okc, int robot_id)
{
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("_okc_send_exit_gracefully: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->command[robot_id] = OKC_SET_EXIT_GRACEFULLY;
  okc->send_break[robot_id] = 1;
  
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("_okc_send_exit_gracefully: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}
  
/**
 * This will stop the server thread and send an "exit gracefully" to
 * all attached robots. The KRL programms on the attached robot(s)
 * should terminate nicely. When function returns, server thread has
 * terminated. Everything should be nice and clean then.
 *
 * @param okc Pointer to #okc_handle_t.  
 * 
 * @return always OKC_OK.
 * @see okc_start_server
 */


int
okc_stop_server (okc_handle_t* okc)
{
  struct timespec ts;
  int i;

  for (i = 0; i < OKC_MAX_ROBOTS; i++)
    {
      if (OKC_OK == okc_is_robot_avail (okc,i))
	{
	  _okc_send_exit_gracefully (okc,i);
	}
    }
  ts.tv_sec = 1;
  ts.tv_nsec = 0;
      
  nanosleep (&ts,NULL);
  
  for (i = 0; i < OKC_MAX_ROBOTS; i++)
    {
      if (OKC_OK == okc_is_robot_avail (okc,i))
	{
	  okc_send_break_to_robot (okc,i);
	}
    }

  ts.tv_sec = 1;
  ts.tv_nsec = 0;
      
  nanosleep (&ts,NULL);
  
  okc->flags = OKC_QUIT;

  pthread_join ((*okc->comm_thread), NULL);

  okc_destroy_okc_handle_t (okc);

  return (OKC_OK);
}

/**
 * This will start the server thread. The server will from then on
 * listen to the given port for robots that are connecting. When a
 * robot connects, the server thread will take care of answering in
 * time.
 *
 * Basically, there are two possible operation modes. Either, you
 * choose OKC_MODE_SETTER. In this mode, all corrections must be done
 * using okc_set_axis_correction() or okc_set_coords_correction(). You
 * can mix calls to these functions, but you should not use both at
 * the same time, that is, you should set axis corrections to 0.0 when
 * using coords correction and vice versa. When using this mode, you
 * will have to take care of "syncing" with the IPO-Cycle of the robot
 * in order to do propper control. The use of a timer can help here,
 * to avoid "busy waiting". When in this mode, refresh_correction and
 * priv_for_refresh_correction are not used an should be set to #NULL.
 *
 * The alternative is to register a callback-function and set either
 * #OKC_MODE_CALLBACK_AXIS or #OKC_MODE_CALLBACK_POS. In these modes,
 * you have to register a callback function (refresh_correction),
 * which will get the priv_for_refresh_correction as first argument
 * (to access custom objects from within the callback function, for
 * example, certain regions in the memory where a trajectorie planner
 * puts the desired values. The next two parameters depend on the
 * mode. In _AXIS mode, a pointer to a double array of size 7, which
 * may be casted to a lbr_axis struct, is passed holding the actual
 * axis positions, and a pointer to a double array of the same length
 * is passed, where the callback function should write the new
 * corrections to. In _POS mode, the array size is 6, and the double*
 * may be casted to a coords struct. When in callback mode, one should
 * consider, that the server thread calls the callback after receiving
 * an XML message from the robot but before an answer is generated and
 * send. So, the callback routine should be "fast", otherwise, you
 * will feel the pain of soft realtime.
 *
 * @param host IP adress or hostname to bind to (in case, you have more than 
 *        one IP, which might be useful when serving the robots).
 * @param port Port to listen to. Per default, this should be 6075 in OpenKC.
 * @param mode Can be either OKC_MODE_SETTER or OKC_MODE_CALLBACK_*
 * @param refresh_correction Pointer to callback function for the 
 *        OKC_MODE_CALLBACK_* modes.
 * @param priv_for_refresh_correction xxx
 * 
 * @return Pointer to #okc_handle_t.
 * @see okc_start_server
 * @see okc_set_coords_correction
 * @see okc_set_axis_correction
 * @see okc_get_coords_act
 * @see okc_get_axis_act
 */


okc_handle_t* 
okc_start_server (const char* host,const  char* port, int mode)
{
  okc_handle_t* okch;
  int listener; 
  int error;

  int yes=1;       
  int rv;
  struct addrinfo hints, *ai, *p;

  pthread_attr_t thread_attr;
  struct sched_param param;
  int policy,result;


  LIBXML_TEST_VERSION;

  xmlInitParser();

  okch = okc_malloc_init_okc_handle_t();
  if (NULL == okch)
    {
      fprintf (stderr,"okc_start_server: error: could not create okc_handle,"
	       " giving up!\n");
      return (NULL);
    }

  okch->mode = mode;
  
  memset(&hints, 0, sizeof hints);
  hints.ai_family = AF_INET;
  hints.ai_socktype = SOCK_STREAM;
  

  if ((rv = getaddrinfo(host, port, &hints, &ai)) != 0) 
    {
      fprintf(stderr, "okc_start_server: %s\n", 
	      gai_strerror(rv));
      okc_destroy_okc_handle_t (okch);
      return (NULL);
    }
    
  for(p = ai; p != NULL; p = p->ai_next) 
    {
      listener = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
      if (listener < 0) 
	{ 
	  perror ("okc_start_server: socket");
	  continue;
	}
      
      setsockopt(listener, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int));
      
      if (bind(listener, p->ai_addr, p->ai_addrlen) < 0) 
	{
	  perror ("okc_start_server: bind");
	  close(listener);
	  continue;
	}
      
      break;
    }

  if (p == NULL) 
    {
      fprintf(stderr, "okc_start_server: failed to bind\n");
      okc_destroy_okc_handle_t (okch);    
      return (NULL);
    }

  freeaddrinfo(ai); 


  if (listen(listener, 10) == -1) 
    {
      perror("okc_start_server: listen");
      okc_destroy_okc_handle_t (okch);   
      return (NULL);
    }

  if (0 != pthread_attr_init (&thread_attr))
    {
      perror ("okc_start_server: pthread_attr_init");
      okc_destroy_okc_handle_t (okch);
      return (NULL);
    }

  if (0 != pthread_attr_setschedpolicy (&thread_attr, SCHED_FIFO))
    {
      fprintf (stderr,"okc_start_server: warning: no SCHED_FIFO available\n");
    }
  
  if (0 != pthread_attr_setinheritsched (&thread_attr, PTHREAD_EXPLICIT_SCHED))
    {
      fprintf (stderr,"okc_start_server: warning: requesting "
	       "PTHREAD_EXPLICIT_SCHED failed\n");
    }

  param.sched_priority = 60;
  if (0 != pthread_attr_setschedparam(&thread_attr, &param))
    {
      fprintf (stderr,"okc_start_server: warning: no SCHED_PRIORITY available\n");
    }

/*   if (0 != pthread_attr_getschedparam(&thread_attr, &param)) */
/*     { */
/*       fprintf (stderr,"okc_start_server: warning: unable to get SCHED_PRIORITY\n"); */

/*     } */
/*   else */
/*     { */
/*       fprintf (stderr,"okc_start_server: info: SCHED_PRIORITY = %d\n",param.sched_priority); */
/*     } */
  
  okch->listener = listener;

  if (0 != (error = pthread_create (okch->comm_thread, &thread_attr, 
				    _serve_connections, (void*) okch)))
    {
      if (error == EPERM)
	{
	  printf ("okc_start_server: warning: could not create realtime "
		  "priority "
		  "thread: reason: permissions. Will create normal priority "
		  "thread.\n");
	  if (0 != (error = pthread_create (okch->comm_thread, NULL, 
				    _serve_connections, (void*) okch)))
	    {
	      printf ("okc_start_server: error: could not create thread: %s",
		      (error == EAGAIN)?"EAGAIN":"EINVAL");
	      okc_destroy_okc_handle_t (okch);
	      pthread_attr_destroy (&thread_attr);
	      return (NULL);
	    }
	}
      else
	{
	  printf ("okc_start_server: error: could not create thread: %s",
		      (error == EAGAIN)?"EAGAIN":"EINVAL");
	  okc_destroy_okc_handle_t (okch);
	  pthread_attr_destroy (&thread_attr);
	  return (NULL);
	}
    }

  if (0 != (result = pthread_getschedparam (*okch->comm_thread,&policy,
					 &param)))
    {
      perror ("okc_start_server: pthread_getschedparam");
      fprintf (stderr,"result was %d\n",result);
    }
  else
    {
      fprintf (stderr,"okc_start_server: info: thread is in class %s at priority %d\n",
	       (policy==SCHED_OTHER)?"SCHED_OTHER":(policy==SCHED_FIFO)?"SCHED_FIFO":"SCHED_RR", param.sched_priority);
    }

  if (policy != SCHED_FIFO)
    {
      fprintf (stderr,"okc_start_server: info: altering thread priority\n");
      policy = SCHED_FIFO;
      param.sched_priority = 50;
      if (0 != pthread_setschedparam (*okch->comm_thread,policy,&param))
	{
	  perror ("okc_start_server: pthread_setschedparam");
	  fprintf (stderr,"okc_start_server: warning: failed to set up realtime "
		   "task\n");
	}
      
      if (0 != (result = pthread_getschedparam (*okch->comm_thread,&policy,
						&param)))
	{
	  perror ("okc_start_server: pthread_getschedparam");
	  fprintf (stderr,"result was %d\n",result);
	}
      else
	{
	  fprintf (stderr,"okc_start_server: info: thread is in class %s at priority %d\n",
		   (policy==SCHED_OTHER)?"SCHED_OTHER":(policy==SCHED_FIFO)?"SCHED_FIFO":"SCHED_RR", param.sched_priority);
	}
    }

  pthread_attr_destroy (&thread_attr);

  return okch;

}


int
okc_register_callback (okc_handle_t* okc, int robot_id, okc_callback callback, 
		       void* private_pointer_for_callback)
{
  
  if (0 != pthread_rwlock_wrlock (&okc->callback_lock[robot_id]))
    {
      perror ("okc_register_callback: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }

  okc->update_correction_callback[robot_id] = callback;
  okc->private_to_update_correction_callback[robot_id] = 
    private_pointer_for_callback;
  
  if (0 != pthread_rwlock_unlock (&okc->callback_lock[robot_id]))
    {
      perror ("okc_register_callback: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return OKC_OK;
}

/**
 * This function will make the KRL program end the RSI-XML guided
 * movement and perform a PTP movement to the given position. This
 * function will block during the movement, i.e. when the function
 * returns, the movement will have finished. Since RSI-XML
 * communication is alive during the movement, you may still get fresh
 * informations from the robot through the getter functions from a
 * different thread. When function returns, robot is again in RSI-XML
 * guided motion.
 *
 * @param okc Pointer to #okc_handle_t.
 * @param robot_id Robot ID to use.  
 * @param position coords to drive the robot to.
 *
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 * @see okc_send_drive_ptp_axis
 */



int
okc_send_drive_ptp_pos (okc_handle_t* okc, int robot_id, 
			 coords position)
{
  struct timespec ts;
  int status=0;
  
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_send_drive_ptp_pos: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_send_drive_ptp_pos: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->params[robot_id][0] = position.x;
  okc->params[robot_id][1] = position.y;
  okc->params[robot_id][2] = position.z;
  okc->params[robot_id][3] = position.a;
  okc->params[robot_id][4] = position.b;
  okc->params[robot_id][5] = position.c;
  okc->params[robot_id][6] = 0.0;
  okc->send_break[robot_id] = 1;

  okc->command[robot_id] = OKC_DRIVE_PTP;
  okc->subcommand[robot_id] = OKC_DRIVE_PTP_XYZ;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_send_drive_ptp_pos: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  while (status != OKC_ACK_DRIVE_PTP_XYZ)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_send_drive_ptp_pos: could not retrieve robots"
		   " status\n");
	  return (OKC_ERROR);
	}
    }
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_send_drive_ptp_pos: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif
  
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->subcommand[robot_id] = OKC_NORMAL_OP;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;
  okc->send_break[robot_id] = 0;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_send_drive_ptp_pos: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}

/**
 * This function will make the KRL program end the RSI-XML guided
 * movement and perform a PTP movement to the given axis angles. This
 * function will block during the movement, i.e. when the function
 * returns, the movement will have finished. Since RSI-XML
 * communication is alive during the movement, you may still get fresh
 * informations from the robot through the getter functions from a
 * different thread. When function returns, robot is again in RSI-XML
 * guided motion.
 *
 * @param okc Pointer to #okc_handle_t.
 * @param robot_id Robot ID to use.  
 * @param position coords to drive the robot to.
 *
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 * @see okc_send_drive_ptp_pos
 */

int
okc_send_drive_ptp_axis (okc_handle_t* okc, int robot_id, 
			 lbr_axis axis)
{
  struct timespec ts;
  int status=0;
  int i;
  
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_send_drive_ptp_axis: invalid/unavailable robot_id\n");
      return (OKC_UNAVAILABLE);
    }
  for (i=0; i < 100; i++)
    {
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,
		   "okc_send_drive_ptp_axis: unable to retrieve status\n");
	  return (OKC_ERROR);
	}
      if (OKC_OK == status)
	break;
      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
    }

  fprintf (stderr,"checks done\n");

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_send_drive_ptp_axis: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  

  okc->params[robot_id][0] = axis.a1;
  okc->params[robot_id][1] = axis.a2;
  okc->params[robot_id][2] = axis.a3;
  okc->params[robot_id][3] = axis.a4;
  okc->params[robot_id][4] = axis.a5;
  okc->params[robot_id][5] = axis.a6;
  okc->params[robot_id][6] = axis.e1;
  okc->send_break[robot_id] = 1;

  okc->command[robot_id] = OKC_DRIVE_PTP;
  okc->subcommand[robot_id] = OKC_DRIVE_PTP_AXIS;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_send_drive_ptp_axis: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  fprintf (stderr,"command submitted\n");

  while (status != OKC_ACK_DRIVE_PTP_AXIS)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      nanosleep (&ts,NULL);
      
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_send_drive_ptp_axis: could not retrieve robots"
		   " status\n");
	  return (OKC_ERROR);
	}
    }

  fprintf (stderr,"received status\n");

  ts.tv_sec = 0;
  ts.tv_nsec = 3*OKC_NANOSLEEP_CYCLE;
  nanosleep (&ts,NULL);

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_send_drive_ptp_axis: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif
  
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->subcommand[robot_id] = OKC_NORMAL_OP;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;
  okc->send_break[robot_id] = 0;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_send_drive_ptp_axis: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}


/**
 * With this function, the parameters for a call to DESIREDFORCE (see
 * Kuka documentation) can be transfered to local variables of the KRL
 * program. These parameters do not take effect until
 * okc_enable_desiredforce() is called. Please note, that some of
 * these parameters are profile dependant, which means, that some
 * parameters have different meanings depending on the kind of
 * forceprofile (Impulse vs. sine). For details, consult Kuka manual.
 *
 * @param okc Pointer to #okc_handle_t.
 * @param robot_id Robot ID to use.  
 * @param forceprofile Can be either #OKC_DESIREDFORCE_IMPULS or 
 *        #OKC_DESIREDFORCE_SINE.
 * @param forcetiming Can be either #OKC_DESIREDFORCE_SYNC or 
 *        #OKC_DESIREDFORCE_ASYNC.
 * @param amplitude Amplitude in Newton (for translational forces) or in 
 *        Newtonmeter for rotational forces.
 * @param duration Duration of how long the force should be applied in seconds.
 * @param max_displacement Defines the maximal displacements in mm or degrees.
 * @param max_velocity Defines the maximal velocity in m/s or degrees/s.
 * @param profile_dependant1 If in #OKC_DESIREDFORCE_IMPULS, it defines the time
 *        of the attack and decay ramps. If #OKC_DESIREDFORCE_SINE is set, it 
 *        defines the frequency of the sine (max. 15 Hz) in Hz. See Kuka 
 *        documentation for details.
 * @param profile_dependant2 Only used if #OKC_DESIREDFORCE_SINE is set. Sets the
 *        phase of the sine in degrees.
 * @param profile_dependant3 Only used if #OKC_DESIREDFORCE_SINE is set. Sets the
 *        force offset with respect to the amplitude.
 *
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 * @see okc_enable_desiredforce
 * @see okc_disable_desiredforce
 */


int
okc_set_desiredforce_params (okc_handle_t* okc, int robot_id, 
			      int forceprofile, int forcetiming, float amplitude,
			      float duration, float max_displacement,
			      float max_velocity, float profile_dependant1,
			      float profile_dependant2, float profile_dependant3)
{
  struct timespec ts;
  int status=0;
  
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_set_desiredforce_params: invalid/unavailable robot_id\n");
      return (OKC_ERROR);
    }

  if (OKC_OK != okc_get_status (okc,robot_id,&status))
    {
      fprintf (stderr,"okc_set_desiredforce_params: could not retrieve robots"
	       " status\n");
      return (OKC_ERROR);
    }

  while (status != OKC_OK)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      fprintf (stderr,"waiting on pending command, status = %d\n",status);
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_set_desiredforce_params: could not retrieve"
		   " robots status\n");
	  return (OKC_ERROR);
	}
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_desiredforce_params: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->forceprofile[robot_id] = forceprofile;
  /* we are using DESIREDFORCESTART anyways... */
  okc->forceimmediate[robot_id] = OKC_DESIREDFORCE_REQUIRE_DESIREDFORCESTART; 
  okc->forcetiming[robot_id] = forcetiming;
  okc->params[robot_id][0] = amplitude;
  okc->params[robot_id][1] = duration;
  okc->params[robot_id][2] = max_displacement;
  okc->params[robot_id][3] = max_velocity;
  okc->params[robot_id][4] = profile_dependant1;
  okc->params[robot_id][5] = profile_dependant2;
  okc->params[robot_id][6] = profile_dependant3;
  okc->command[robot_id] = OKC_SET_DESIREDFORCE_PARAMS;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_desiredforce_params: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  while (status != OKC_ACK_SET_DESIREDFORCE_PARAMS)
    {
      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      nanosleep (&ts,NULL);
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_set_desiredforce_params: could not "
		   "retrieve robots status\n");
	  return (OKC_ERROR);
	}
        }
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_desiredforce_params: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif
  
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;
  okc->send_break[robot_id] = 0;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_set_desiredforce_params: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  return (OKC_OK);
}

/**
 * Function intern to OpenKC. Checks, if direction is valid.
 *
 * @param dir Direction to check.
 * 
 * @return OKC_OK, if dir is in OKC_?_AXIS, OKC_ERROR elsewise.
 */

static int
_check_direction (int dir)
{
  
  if (dir == OKC_X_AXIS ||
      dir == OKC_Y_AXIS ||
      dir == OKC_Z_AXIS)
    return (OKC_OK);
  return (OKC_ERROR);
}
  

/**
 * A call to this function will enable the DESIREDFORCE controller (or
 * update the controller). The parameters must have been previously
 * set by a call to okc_set_desiredforce_params(). When the function
 * returns, the DESIREDFORE-contruct will be enabled on the
 * robot. Please note, that you should be in cartesian impedance
 * control, if you want to "see" something.
 *
 * @param okc Pointer to #okc_handle_t.
 * @param robot_id Robot ID to use.  
 * @param direction Either #OKC_X_AXIS, #OKC_Y_AXIS or #OKC_Z_AXIS. Force
 *                  will be enabled in that direction. 
 *
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 * @see okc_set_desiredforce_params
 * @see okc_disable_desiredforce
 */


int
okc_enable_desiredforce (okc_handle_t* okc, int robot_id, int direction)
{
  struct timespec ts;
  int status=0;
  int ack=0;

  if (OKC_OK != _check_direction (direction))
    {
      fprintf (stderr,"okc_enable_desiredforce: invalid parameter:"
	       "direction (is %d)\n",direction);
      return (OKC_UNAVAILABLE);
    }
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_enable_desiredforce: invalid/unavailable "
	       "robot_id\n");
      return (OKC_ERROR);
    }
  if (0 != pthread_rwlock_rdlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_enable_desiredforce: pthread_rwlock_rdlock");
      return (OKC_ERROR);
    }
  
  if (okc->forceimmediate[robot_id] < 0)
    {
      fprintf (stderr,
	       "okc_enable_desiredforce: trying to make force active"
	       "without havine set parameters first!\n");
      ack=OKC_ERROR;
    }
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_enable_desiredforce: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  if (OKC_ERROR == ack)
    return (OKC_ERROR);

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_enable_desiredforce: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->command[robot_id] = OKC_SWITCH_ON_DESIREDFORCE;
  switch (direction)
    {
    case OKC_X_AXIS: 
      okc->subcommand[robot_id] = OKC_SET_DESIREDFORCE_X;
      ack = OKC_ACK_SET_DESIREDFORCE_X;
      break;
    case OKC_Y_AXIS: 
      okc->subcommand[robot_id] = OKC_SET_DESIREDFORCE_Y;
      ack = OKC_ACK_SET_DESIREDFORCE_Y;
      break;
    case OKC_Z_AXIS:
      okc->subcommand[robot_id] = OKC_SET_DESIREDFORCE_Z;
      ack = OKC_ACK_SET_DESIREDFORCE_Z;
      break;
    }
  okc->send_break[robot_id] = 1;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_enable_desiredforce: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  while (status != ack)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_enable_desiredforce: could not "
		   "retrieve robots status\n");
	  return (OKC_ERROR);
	}
    }
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_enable_desiredforce: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif  
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->subcommand[robot_id] = OKC_NORMAL_OP;
  okc->send_break[robot_id] = 0;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;
  
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_enable_desiredforce: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}


/**
 * A call to this function will disable the DESIREDFORCE controller. 
 *
 * @param okc Pointer to #okc_handle_t.
 * @param robot_id Robot ID to use.  
 *

 *
 * @return OKC_OK, if setting was successful, OKC_UNAVAILABLE
 *         if robot is unavailable (disconnected) or OKC_ERROR on error.
 * @see okc_set_desiredforce_params
 * @see okc_enable_desiredforce
 */

int
okc_disable_desiredforce (okc_handle_t* okc, int robot_id)
{
  struct timespec ts;
  int status=0;
  
  if (OKC_OK != okc_is_robot_avail (okc,robot_id))
    {
      fprintf (stderr,
	       "okc_disable_desiredforce: invalid/unavailable "
	       "robot_id\n");
      return (OKC_UNAVAILABLE);
    }

  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_disable_desiredforce: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
  
  okc->command[robot_id] = OKC_SWITCH_OFF_DESIREDFORCE;
  okc->send_break[robot_id] = 1;

  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_disable_desiredforce: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }

  while (status != OKC_ACK_SWITCH_OFF_DESIREDFORCE)
    {

      ts.tv_sec = 0;
      ts.tv_nsec = OKC_NANOSLEEP_CYCLE;
      
      nanosleep (&ts,NULL);
      
      if (OKC_OK != okc_get_status (okc,robot_id,&status))
	{
	  fprintf (stderr,"okc_disable_desiredforce: could not "
		   "retrieve robots status\n");
	  return (OKC_ERROR);
	}
    }
  
  if (0 != pthread_rwlock_wrlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_disable_desiredforce: pthread_rwlock_wrlock");
      return (OKC_ERROR);
    }
#ifdef TICKS_DEBUG
  fprintf (stderr,"Command %d took %d ticks\n",okc->pending_command[robot_id],
	   okc->ticks_command[robot_id]);
#endif  
  okc->command[robot_id] = OKC_NORMAL_OP;
  okc->send_break[robot_id] = 0;
  okc->ticks_command[robot_id] = 0;
  okc->pending_command[robot_id] = OKC_NORMAL_OP;
  
  if (0 != pthread_rwlock_unlock (&okc->pos_axis_in_lock[robot_id]))
    {
      perror ("okc_disable_desiredforce: pthread_rwlock_unlock");
      return (OKC_ERROR);
    }
  return (OKC_OK);
}
